/* bz-result.h
 *
 * Copyright 2025 Adam Masciola
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <libdex.h>

G_BEGIN_DECLS

#define BZ_TYPE_RESULT (bz_result_get_type ())
G_DECLARE_FINAL_TYPE (BzResult, bz_result, BZ, RESULT, GObject)

BzResult *
bz_result_new (DexFuture *future);

gboolean
bz_result_get_pending (BzResult *self);

gboolean
bz_result_get_resolved (BzResult *self);

gboolean
bz_result_get_rejected (BzResult *self);

gpointer
bz_result_get_object (BzResult *self);

const char *
bz_result_get_message (BzResult *self);

DexFuture *
bz_result_dup_future (BzResult *self);

G_END_DECLS

/* End of bz-result.h */
